@@ -1,45 +1,83 @@
 use std::collections::{BTreeMap, BTreeSet};
 
-use blsttc::{PublicKeyShare, SecretKeyShare};
+use blsttc::{PublicKeySet, SecretKeyShare, Signature, SignatureShare};
 use log::info;
-use rand::{CryptoRng, Rng};
+use serde::Serialize;
 
 use crate::sn_membership::Generation;
 use crate::vote::{Ballot, Proposition, SignedVote, Vote};
-use crate::{Error, Result};
+use crate::{Error, NodeId, Result};
 
 #[derive(Debug)]
 pub struct Consensus<T: Proposition> {
-    pub elders: BTreeSet<PublicKeyShare>,
-    pub secret_key: SecretKeyShare,
-    pub votes: BTreeMap<PublicKeyShare, SignedVote<T>>,
+    pub elders: PublicKeySet,
+    pub n_elders: usize,
+    pub secret_key: (NodeId, SecretKeyShare),
+    pub votes: BTreeMap<NodeId, SignedVote<T>>,
 }
 
 pub enum VoteResponse<T: Proposition> {
     WaitingForMoreVotes,
     Broadcast(SignedVote<T>),
-    Decided(SignedVote<T>),
+    Decided {
+        votes: BTreeSet<SignedVote<T>>,
+        proposals: BTreeMap<T, Signature>,
+    },
 }
 
 impl<T: Proposition> Consensus<T> {
-    pub fn from(secret_key: SecretKeyShare, elders: BTreeSet<PublicKeyShare>) -> Self {
+    pub fn from(
+        secret_key: (NodeId, SecretKeyShare),
+        elders: PublicKeySet,
+        n_elders: usize,
+    ) -> Self {
         Consensus::<T> {
-            secret_key,
             elders,
+            n_elders,
+            secret_key,
             votes: Default::default(),
         }
     }
 
-    pub fn random(mut rng: impl Rng + CryptoRng) -> Self {
-        Consensus::<T> {
-            secret_key: rng.gen(),
-            elders: Default::default(),
-            votes: Default::default(),
+    pub fn verify_sig_share<M: Serialize>(
+        &self,
+        msg: &M,
+        elder: NodeId,
+        sig: &SignatureShare,
+    ) -> Result<()> {
+        let public_key = self.elders.public_key_share(elder as u64);
+        let msg_bytes = bincode::serialize(msg)?;
+        if public_key.verify(sig, msg_bytes) {
+            Ok(())
+        } else {
+            Err(Error::InvalidElderSignature)
         }
     }
 
-    pub fn public_key_share(&self) -> PublicKeyShare {
-        self.secret_key.public_key_share()
+    pub fn sign<M: Serialize>(&self, msg: &M) -> Result<SignatureShare> {
+        Ok(self.secret_key.1.sign(&bincode::serialize(msg)?))
+    }
+
+    pub fn id(&self) -> NodeId {
+        self.secret_key.0
+    }
+
+    pub fn build_super_majority_vote(
+        &self,
+        votes: BTreeSet<SignedVote<T>>,
+        gen: Generation,
+    ) -> Result<SignedVote<T>> {
+        let proposals: BTreeMap<T, (NodeId, SignatureShare)> = self
+            .proposals(&votes)
+            .into_iter()
+            .map(|p| {
+                let sig = self.sign(&p)?;
+                Ok((p, (self.secret_key.0, sig)))
+            })
+            .collect::<Result<_>>()?;
+        let ballot = Ballot::SuperMajority { votes, proposals }.simplify();
+        let vote = Vote { gen, ballot };
+        self.sign_vote(vote)
     }
 
     // handover: gen = gen
@@ -53,7 +91,7 @@ impl<T: Proposition> Consensus<T> {
     ) -> Result<VoteResponse<T>> {
         self.log_signed_vote(&signed_vote);
 
-        if self.is_split_vote(&self.votes.values().cloned().collect())? {
+        if self.is_split_vote(&self.votes.values().cloned().collect()) {
             info!("[MBR] Detected split vote");
             let merge_vote = Vote {
                 gen,
@@ -61,14 +99,9 @@ impl<T: Proposition> Consensus<T> {
             };
             let signed_merge_vote = self.sign_vote(merge_vote)?;
 
-            if let Some(our_vote) = self.votes.get(&self.public_key_share()) {
-                let proposals_we_voted_for =
-                    BTreeSet::from_iter(our_vote.proposals().into_iter().map(|(_, r)| r));
-                let proposals_we_would_vote_for: BTreeSet<_> = signed_merge_vote
-                    .proposals()
-                    .into_iter()
-                    .map(|(_, r)| r)
-                    .collect();
+            if let Some(our_vote) = self.votes.get(&self.id()) {
+                let proposals_we_voted_for = our_vote.proposals();
+                let proposals_we_would_vote_for = signed_merge_vote.proposals();
 
                 if proposals_we_voted_for == proposals_we_would_vote_for {
                     info!("[MBR] This vote didn't add new information, waiting for more votes...");
@@ -80,23 +113,20 @@ impl<T: Proposition> Consensus<T> {
             return Ok(VoteResponse::Broadcast(self.cast_vote(signed_merge_vote)));
         }
 
-        if self.is_super_majority_over_super_majorities(&self.votes.values().cloned().collect())? {
-            info!("[MBR] Detected super majority over super majorities");
-            assert!(self.elders.contains(&self.public_key_share()));
-            // store a proof of what the network decided in our history so that we can onboard future procs.
-            let ballot = Ballot::SuperMajority(self.votes.values().cloned().collect()).simplify();
-
-            let vote = Vote { gen, ballot };
-            let signed_vote = self.sign_vote(vote)?;
-
-            // return obtained super majority over super majority (aka consensus)
-            return Ok(VoteResponse::Decided(signed_vote));
+        if let Some(proposals) =
+            self.get_super_majority_over_super_majorities(&self.votes.values().cloned().collect())?
+        {
+            info!("[MBR] Detected super majority over super majorities: {proposals:?}");
+            return Ok(VoteResponse::Decided {
+                votes: self.votes.values().cloned().collect(),
+                proposals,
+            });
         }
 
-        if self.is_super_majority(&self.votes.values().cloned().collect())? {
+        if self.is_super_majority(&self.votes.values().cloned().collect()) {
             info!("[MBR] Detected super majority");
 
-            if let Some(our_vote) = self.votes.get(&self.public_key_share()) {
+            if let Some(our_vote) = self.votes.get(&self.id()) {
                 // We voted during this generation.
 
                 if our_vote.vote.is_super_majority_ballot() {
@@ -106,15 +136,14 @@ impl<T: Proposition> Consensus<T> {
             }
 
             info!("[MBR] broadcasting super majority");
-            let ballot = Ballot::SuperMajority(self.votes.values().cloned().collect()).simplify();
-            let vote = Vote { gen, ballot };
-            let signed_vote = self.sign_vote(vote)?;
+            let signed_vote =
+                self.build_super_majority_vote(self.votes.values().cloned().collect(), gen)?;
             return Ok(VoteResponse::Broadcast(self.cast_vote(signed_vote)));
         }
 
         // We have determined that we don't yet have enough votes to take action.
         // If we have not yet voted, this is where we would contribute our vote
-        if !self.votes.contains_key(&self.public_key_share()) {
+        if !self.votes.contains_key(&self.id()) {
             let signed_vote = self.sign_vote(Vote {
                 gen,
                 ballot: signed_vote.vote.ballot,
@@ -127,8 +156,8 @@ impl<T: Proposition> Consensus<T> {
 
     pub fn sign_vote(&self, vote: Vote<T>) -> Result<SignedVote<T>> {
         Ok(SignedVote {
-            voter: self.public_key_share(),
-            sig: self.secret_key.sign(&vote.to_bytes()?),
+            voter: self.secret_key.0,
+            sig: self.sign(&vote)?,
             vote,
         })
     }
@@ -151,7 +180,7 @@ impl<T: Proposition> Consensus<T> {
         let mut count: BTreeMap<BTreeSet<T>, usize> = Default::default();
 
         for vote in votes.iter() {
-            let proposals = BTreeSet::from_iter(vote.proposals().into_iter().map(|(_, prop)| prop));
+            let proposals = vote.proposals();
             let c = count.entry(proposals).or_default();
             *c += 1;
         }
@@ -159,22 +188,23 @@ impl<T: Proposition> Consensus<T> {
         count
     }
 
-    fn is_split_vote(&self, votes: &BTreeSet<SignedVote<T>>) -> Result<bool> {
+    fn proposals(&self, votes: &BTreeSet<SignedVote<T>>) -> BTreeSet<T> {
+        BTreeSet::from_iter(votes.iter().flat_map(|v| v.proposals()))
+    }
+
+    fn is_split_vote(&self, votes: &BTreeSet<SignedVote<T>>) -> bool {
         let counts = self.count_votes(votes);
         let most_votes = counts.values().max().cloned().unwrap_or_default();
         let voters = BTreeSet::from_iter(votes.iter().map(|v| v.voter));
-        let remaining_voters = self.elders.difference(&voters).count();
+        let remaining_voters = self.n_elders - voters.len();
 
         // give the remaining votes to the proposals with the most votes.
         let predicted_votes = most_votes + remaining_voters;
 
-        Ok(
-            3 * voters.len() > 2 * self.elders.len()
-                && 3 * predicted_votes <= 2 * self.elders.len(),
-        )
+        voters.len() > self.elders.threshold() && predicted_votes <= self.elders.threshold()
     }
 
-    pub fn is_super_majority(&self, votes: &BTreeSet<SignedVote<T>>) -> Result<bool> {
+    pub fn is_super_majority(&self, votes: &BTreeSet<SignedVote<T>>) -> bool {
         // TODO: super majority should always just be the largest 7 members
         let most_votes = self
             .count_votes(votes)
@@ -183,36 +213,59 @@ impl<T: Proposition> Consensus<T> {
             .cloned()
             .unwrap_or_default();
 
-        Ok(3 * most_votes > 2 * self.elders.len())
+        most_votes > self.elders.threshold()
     }
 
-    fn is_super_majority_over_super_majorities(
+    fn get_super_majority_over_super_majorities(
         &self,
         votes: &BTreeSet<SignedVote<T>>,
-    ) -> Result<bool> {
-        let count_of_agreeing_super_majorities = self
-            .count_votes(&BTreeSet::from_iter(
-                votes
-                    .iter()
-                    .filter(|v| v.vote.is_super_majority_ballot())
-                    .cloned(),
-            ))
+    ) -> Result<Option<BTreeMap<T, Signature>>> {
+        let sm_votes = BTreeSet::from_iter(
+            votes
+                .iter()
+                .filter(|v| v.vote.is_super_majority_ballot())
+                .cloned(),
+        );
+        if let Some((props, count_of_sm)) = self
+            .count_votes(&sm_votes)
             .into_iter()
-            .map(|(_, count)| count)
-            .max()
-            .unwrap_or(0);
+            .max_by_key(|(_, c)| *c)
+        {
+            if count_of_sm > self.elders.threshold() {
+                let mut proposal_sigs: BTreeMap<T, BTreeSet<(u64, SignatureShare)>> =
+                    Default::default();
+                let proposals_from_each_supermajority = sm_votes
+                    .iter()
+                    .filter(|sm| sm.proposals() == props)
+                    .flat_map(|v| match &v.vote.ballot {
+                        Ballot::SuperMajority { proposals, .. } => proposals.clone(),
+                        _ => Default::default(),
+                    });
+                for (prop, (id, sig)) in proposals_from_each_supermajority {
+                    proposal_sigs
+                        .entry(prop)
+                        .or_default()
+                        .insert((id as u64, sig));
+                }
 
-        Ok(3 * count_of_agreeing_super_majorities > 2 * self.elders.len())
+                return Ok(Some(
+                    proposal_sigs
+                        .into_iter()
+                        .map(|(prop, sigs)| Ok((prop, self.elders.combine_signatures(sigs)?)))
+                        .collect::<Result<_>>()?,
+                ));
+            }
+        }
+
+        Ok(None)
     }
 
     /// Validates a vote recursively all the way down to the proposition (T)
     /// Assumes those propositions are correct, they MUST be checked beforehand by the caller
     pub fn validate_signed_vote(&self, signed_vote: &SignedVote<T>) -> Result<()> {
-        signed_vote.validate_signature()?;
+        self.verify_sig_share(&signed_vote.vote, signed_vote.voter, &signed_vote.sig)?;
         self.validate_vote(&signed_vote.vote)?;
-        self.validate_is_elder(signed_vote.voter)?;
         self.validate_vote_supersedes_existing_vote(signed_vote)?;
-        self.validate_voters_have_not_changed_proposals(signed_vote)?;
         Ok(())
     }
 
@@ -231,15 +284,23 @@ impl<T: Proposition> Consensus<T> {
                 }
                 Ok(())
             }
-            Ballot::SuperMajority(votes) => {
+            Ballot::SuperMajority { votes, proposals } => {
                 if !self.is_super_majority(
                     &votes
                         .iter()
                         .flat_map(SignedVote::unpack_votes)
                         .cloned()
                         .collect(),
-                )? {
+                ) {
                     Err(Error::SuperMajorityBallotIsNotSuperMajority)
+                } else if vote.proposals() != BTreeSet::from_iter(proposals.keys().cloned()) {
+                    Err(Error::SuperMajorityProposalsDoesNotMatchVoteProposals)
+                } else if proposals
+                    .iter()
+                    .try_for_each(|(p, (id, sig))| self.verify_sig_share(&p, *id, sig))
+                    .is_err()
+                {
+                    Err(Error::InvalidElderSignature)
                 } else {
                     for child_vote in votes.iter() {
                         if child_vote.vote.gen != vote.gen {
@@ -256,17 +317,6 @@ impl<T: Proposition> Consensus<T> {
         }
     }
 
-    fn validate_is_elder(&self, public_key: PublicKeyShare) -> Result<()> {
-        if !self.elders.contains(&public_key) {
-            Err(Error::NotElder {
-                public_key,
-                elders: self.elders.clone(),
-            })
-        } else {
-            Ok(())
-        }
-    }
-
     fn validate_vote_supersedes_existing_vote(&self, signed_vote: &SignedVote<T>) -> Result<()> {
         if self.votes.contains_key(&signed_vote.voter)
             && !signed_vote.supersedes(&self.votes[&signed_vote.voter])
@@ -277,24 +327,4 @@ impl<T: Proposition> Consensus<T> {
             Ok(())
         }
     }
-
-    fn validate_voters_have_not_changed_proposals(
-        &self,
-        signed_vote: &SignedVote<T>,
-    ) -> Result<()> {
-        // Ensure that nobody is trying to change their Proposal proposals.
-        let proposals: BTreeSet<(PublicKeyShare, T)> = self
-            .votes
-            .values()
-            .flat_map(|v| v.proposals())
-            .chain(signed_vote.proposals())
-            .collect();
-
-        let voters = BTreeSet::from_iter(proposals.iter().map(|(actor, _)| actor));
-        if voters.len() != proposals.len() {
-            Err(Error::VoterChangedVote)
-        } else {
-            Ok(())
-        }
-    }
 }
