@@ -1,11 +1,11 @@
-use std::collections::BTreeSet;
+use std::collections::{BTreeMap, BTreeSet};
 
-use blsttc::{PublicKeyShare, SignatureShare};
+use blsttc::SignatureShare;
 use core::fmt::Debug;
 use serde::{Deserialize, Serialize};
 
 use crate::sn_membership::Generation;
-use crate::{Error, Result};
+use crate::NodeId;
 
 pub trait Proposition: Ord + Clone + Debug + Serialize {}
 impl<T: Ord + Clone + Debug + Serialize> Proposition for T {}
@@ -14,7 +14,10 @@ impl<T: Ord + Clone + Debug + Serialize> Proposition for T {}
 pub enum Ballot<T: Proposition> {
     Propose(T),
     Merge(BTreeSet<SignedVote<T>>),
-    SuperMajority(BTreeSet<SignedVote<T>>),
+    SuperMajority {
+        votes: BTreeSet<SignedVote<T>>,
+        proposals: BTreeMap<T, (NodeId, SignatureShare)>,
+    },
 }
 
 impl<T: Proposition> Debug for Ballot<T> {
@@ -22,7 +25,12 @@ impl<T: Proposition> Debug for Ballot<T> {
         match self {
             Ballot::Propose(r) => write!(f, "P({:?})", r),
             Ballot::Merge(votes) => write!(f, "M{:?}", votes),
-            Ballot::SuperMajority(votes) => write!(f, "SM{:?}", votes),
+            Ballot::SuperMajority { votes, proposals } => write!(
+                f,
+                "SM{:?}-{:?}",
+                votes,
+                BTreeSet::from_iter(proposals.keys())
+            ),
         }
     }
 }
@@ -45,11 +53,15 @@ fn simplify_votes<T: Proposition>(
 }
 
 impl<T: Proposition> Ballot<T> {
+    #[must_use]
     pub fn simplify(&self) -> Self {
         match &self {
             Ballot::Propose(_) => self.clone(), // already in simplest form
             Ballot::Merge(votes) => Ballot::Merge(simplify_votes(votes)),
-            Ballot::SuperMajority(votes) => Ballot::SuperMajority(simplify_votes(votes)),
+            Ballot::SuperMajority { votes, proposals } => Ballot::SuperMajority {
+                votes: simplify_votes(votes),
+                proposals: proposals.clone(),
+            },
         }
     }
 }
@@ -67,53 +79,46 @@ impl<T: Proposition> Debug for Vote<T> {
 }
 
 impl<T: Proposition> Vote<T> {
-    pub fn to_bytes(&self) -> Result<Vec<u8>> {
-        Ok(bincode::serialize(&(&self.ballot, &self.gen))?)
+    pub fn is_super_majority_ballot(&self) -> bool {
+        matches!(self.ballot, Ballot::SuperMajority { .. })
     }
 
-    pub fn is_super_majority_ballot(&self) -> bool {
-        matches!(self.ballot, Ballot::SuperMajority(_))
+    pub fn proposals(&self) -> BTreeSet<T> {
+        match &self.ballot {
+            Ballot::Propose(proposal) => BTreeSet::from_iter([proposal.clone()]),
+            Ballot::Merge(votes) | Ballot::SuperMajority { votes, .. } => {
+                // TAI: use proposals instead of recursing on SuperMajority?
+                BTreeSet::from_iter(votes.iter().flat_map(SignedVote::proposals))
+            }
+        }
     }
 }
 
 #[derive(Clone, PartialEq, Eq, PartialOrd, Ord, Serialize, Deserialize)]
 pub struct SignedVote<T: Proposition> {
     pub vote: Vote<T>,
-    pub voter: PublicKeyShare,
+    pub voter: NodeId,
     pub sig: SignatureShare,
 }
 
 impl<T: Proposition> Debug for SignedVote<T> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "{:?}@{:?}", self.vote, self.voter)
+        write!(f, "{:?}@{}", self.vote, self.voter)
     }
 }
 
 impl<T: Proposition> SignedVote<T> {
-    pub fn validate_signature(&self) -> Result<()> {
-        if self.voter.verify(&self.sig, &self.vote.to_bytes()?) {
-            Ok(())
-        } else {
-            Err(Error::InvalidElderSignature)
-        }
-    }
-
     pub fn unpack_votes(&self) -> BTreeSet<&Self> {
         match &self.vote.ballot {
             Ballot::Propose(_) => BTreeSet::from_iter([self]),
-            Ballot::Merge(votes) | Ballot::SuperMajority(votes) => BTreeSet::from_iter(
+            Ballot::Merge(votes) | Ballot::SuperMajority { votes, .. } => BTreeSet::from_iter(
                 std::iter::once(self).chain(votes.iter().flat_map(Self::unpack_votes)),
             ),
         }
     }
 
-    pub fn proposals(&self) -> BTreeSet<(PublicKeyShare, T)> {
-        match &self.vote.ballot {
-            Ballot::Propose(proposal) => BTreeSet::from_iter([(self.voter, proposal.clone())]),
-            Ballot::Merge(votes) | Ballot::SuperMajority(votes) => {
-                BTreeSet::from_iter(votes.iter().flat_map(Self::proposals))
-            }
-        }
+    pub fn proposals(&self) -> BTreeSet<T> {
+        self.vote.proposals()
     }
 
     pub fn supersedes(&self, signed_vote: &Self) -> bool {
@@ -122,7 +127,7 @@ impl<T: Proposition> SignedVote<T> {
         } else {
             match &self.vote.ballot {
                 Ballot::Propose(_) => false,
-                Ballot::Merge(votes) | Ballot::SuperMajority(votes) => {
+                Ballot::Merge(votes) | Ballot::SuperMajority { votes, .. } => {
                     votes.iter().any(|v| v.supersedes(signed_vote))
                 }
             }
